@@ -79,8 +79,14 @@ bool TextSplit::o_processCJK{true};
 int  TextSplit::o_CJKNgramLen{2};
 bool TextSplit::o_noNumbers{false};
 bool TextSplit::o_deHyphenate{false};
+// Max word size in Unicode characters
 int  TextSplit::o_maxWordLength{40};
 int  TextSplit::o_maxWordsInSpan{6};
+// Do we produce spans like 12-34-56?
+bool TextSplit::o_numberDashSpans{false};
+
+static std::unordered_set<unsigned int> o_idxpunct;
+static bool o_haveidxpunct;
 
 static bool o_exthangultagger{false};
 static bool o_extchinesetagger{false};
@@ -92,7 +98,18 @@ void TextSplit::staticConfInit(RclConfig *config)
 {
     config->getConfParam("maxtermlength", &o_maxWordLength);
     config->getConfParam("maxwordsinspan", &o_maxWordsInSpan);
+    config->getConfParam("numberdashspans", &o_numberDashSpans);
 
+    std::string sidxpunct;
+    config->getConfParam("indexedpunctuation", sidxpunct);
+    if (!sidxpunct.empty()) {
+        Utf8Iter it(sidxpunct);
+        for (; !it.eof() && !it.error(); it++) {
+            o_idxpunct.insert(*it);
+        }
+        o_haveidxpunct = true;
+    }
+    
     bool bvalue{false};
     if (config->getConfParam("nocjk", &bvalue) && bvalue == true) {
         o_processCJK = false;
@@ -115,11 +132,11 @@ void TextSplit::staticConfInit(RclConfig *config)
         o_deHyphenate = bvalue;
     }
 
+    // Note: charclasses is already initialized (by a static initializer). \ is set as "special"
     bvalue = false;
     if (config->getConfParam("backslashasletter", &bvalue)) {
         if (bvalue) {
-        } else {
-            charclasses[int('\\')] = SPACE;
+            charclasses[int('\\')] = A_LLETTER;
         }
     }
 
@@ -160,7 +177,8 @@ public:
     CharClassInit() {
         unsigned int i;
 
-        // Set default value for all: SPACE
+        // Set default value for all: SPACE. Only the low 127 slots are used of course, but it seems
+        // that having the full 256 makes some tests simpler (not checked recently).
         for (i = 0 ; i < 256 ; i ++)
             charclasses[i] = TextSplit::SPACE;
 
@@ -194,6 +212,8 @@ public:
         for (i = 0; i  < strlen(special); i++)
             charclasses[int(special[i])] = special[i];
 
+
+        // Non-ASCII Unicode punctuation and white space 
         for (i = 0; i < sizeof(unipunc) / sizeof(int); i++) {
             spunc.insert(unipunc[i]);
         }
@@ -246,7 +266,8 @@ int TextSplit::whatcc(unsigned int c)
     }
 }
 
-// CJK Unicode character detection. CJK text is generally indexed using an n-gram method.
+// Chinese, Japanese, Korean Unicode character ranges. CJK text is generally indexed using an n-gram
+// method.
 // Exceptions:
 //  Hangul: we can use an external text linguistic-aware segmenter.
 //  Chinese: same. Using Jieba
@@ -301,8 +322,8 @@ int TextSplit::whatcc(unsigned int c)
 #define UNICODE_IS_KATAKANA(p) false
 #endif
 
-// Hangul detection and options. If no external tagger is configured, we process HANGUL as generic
-// CJK (n-grams)
+// Hangul ranges and options. If no external tagger is configured, we process HANGUL as generic CJK
+// (n-grams): IS_HANGUL will be false and IS_CJK true.
 #define UNICODE_IN_HANGUL_RANGE(p)             \
         (((p) >= 0x1100 && (p) <= 0x11FF) ||   \
          ((p) >= 0x3130 && (p) <= 0x318F) ||   \
@@ -316,24 +337,58 @@ int TextSplit::whatcc(unsigned int c)
 #define UNICODE_IS_HANGUL(p) false
 #endif
 
-// Same for Chinese
+// Chinese ranges: everything in CJK range defined above which is not Katakana or Hangul
 #define UNICODE_IN_CHINESE_RANGE(p) \
     (UNICODE_IS_CJK(p) && !(UNICODE_IN_KATAKANA_RANGE(p) || UNICODE_IN_HANGUL_RANGE(p)))
 #ifdef CHINESE_AS_WORDS
+// If we are not using an external text splitter, IS_CJK will be true and IS_CHINESE false
 #define UNICODE_IS_CHINESE(p) (o_extchinesetagger && UNICODE_IN_CHINESE_RANGE(p))
 #else
 #define UNICODE_IS_CHINESE(p) false
 #endif
 
+// Tibetan. For now, uses n-grams. Not CJK though so handled a bit differently (because not in the
+// big CJK range)
+#define UNICODE_IN_TIBETAN_RANGE(p) \
+    ((p) > 0x0f00 && (p) <= 0x0fff)
+#define UNICODE_IS_TIBETAN(p) UNICODE_IN_TIBETAN_RANGE(p)
+
+// Thai and Lao range are contiguous
+#define UNICODE_IN_THAI_LAO_RANGE(p) \
+    ((p) > 0x0e00 && (p) <= 0x0eff)
+#define UNICODE_IS_THAI_OR_LAO(p) UNICODE_IN_THAI_LAO_RANGE(p)
+
+// Burmese
+#define UNICODE_IN_BURMESE_RANGE(p) \
+    ((p) > 0x1000 && (p) <= 0x109f)
+#define UNICODE_IS_BURMESE(p) UNICODE_IN_BURMESE_RANGE(p)
+
+// Non CJK scripts processed as ngrams
+#define UNICODE_IS_OTHER_NGRAM(p) (UNICODE_IS_TIBETAN(p) || UNICODE_IS_THAI_OR_LAO(p) || \
+                                   UNICODE_IS_BURMESE(p))
+
+// Just for testing the ngram splitter: this European reader finds it easier to look at results when
+// using a simple A-Z range.
+#define UNICODE_IN_TESTNGRAM_RANGE(p) \
+    ((p) >= 'A' && (p) <= 'B')
+#ifdef TESTING_NGRAMS
+#define UNICODE_IS_TESTNGRAM(p) UNICODE_IN_TESTNGRAM_RANGE(p)
+#else
+#define UNICODE_IS_TESTNGRAM(p) false
+#endif
+
 bool TextSplit::isSpace(int c)
 {
     return whatcc(c) == SPACE;
 }
 bool TextSplit::isCJK(int c)
 {
-    PRETEND_USE(c);
     return UNICODE_IS_CJK(c);
 }
+bool TextSplit::noStemming(int c)
+{
+    return UNICODE_IS_CJK(c) || UNICODE_IS_OTHER_NGRAM(c);
+}
 bool TextSplit::isKATAKANA(int c)
 {
     PRETEND_USE(c);
@@ -351,18 +406,21 @@ bool TextSplit::isCHINESE(int c)
 }
 bool TextSplit::isNGRAMMED(int c)
 {
-    PRETEND_USE(c);
-    return UNICODE_IS_CJK(c) && !UNICODE_IS_KATAKANA(c) && !UNICODE_IS_HANGUL(c);
+    return UNICODE_IS_OTHER_NGRAM(c) || UNICODE_IS_TESTNGRAM(c) || 
+        (UNICODE_IS_CJK(c) && !UNICODE_IS_KATAKANA(c) && !UNICODE_IS_HANGUL(c) &&
+         !UNICODE_IS_CHINESE(c));
 }
 
 
 // This is used to detect katakana/other transitions, which must trigger a word split (there is not
 // always a separator, and katakana is otherwise treated like other, in the same routine, unless CJK
 // which has its span reader causing a word break)
-enum CharSpanClass {CSC_HANGUL, CSC_CHINESE, CSC_CJK, CSC_KATAKANA, CSC_OTHER};
+enum CharSpanClass {CSC_OTHER, CSC_HANGUL, CSC_CHINESE, CSC_CJK, CSC_KATAKANA, CSC_OTHERNGRAM,
+                    CSC_TESTNGRAM};
 std::vector<CharFlags> csc_names {
     CHARFLAGENTRY(CSC_HANGUL), CHARFLAGENTRY(CSC_CHINESE), CHARFLAGENTRY(CSC_CJK),
-    CHARFLAGENTRY(CSC_KATAKANA), CHARFLAGENTRY(CSC_OTHER)};
+    CHARFLAGENTRY(CSC_KATAKANA), CHARFLAGENTRY(CSC_OTHER), CHARFLAGENTRY(CSC_OTHERNGRAM),
+    CHARFLAGENTRY(CSC_TESTNGRAM)};
 
 // Final term checkpoint: do some checking (the kind which is simpler to do here than in the main
 // loop), then send term to our client.
@@ -379,14 +437,18 @@ inline bool TextSplit::emitterm(bool isspan, const string& w, int pos, size_t bt
     PRETEND_USE(isspan);
 #endif
 
-    int l = int(w.length());
-    if (l == 0 || l > o_maxWordLength) {
-        return true;
+    // If the size in bytes is > max size in unicode chars, compute the Unicode length.
+    // The bytes test saves a bit of CPU as most words will skip the unicode test.
+    // In theory we know the character size, but in practise, it was lost somewhere in the maze.
+    if (w.size() == 0 || w.size() >= (size_t)o_maxWordLength) {
+        int l = utf8len(w);
+        if (l == 0 || l > o_maxWordLength) {
+            discarded(w, pos, btstart, btend, WORD_TOO_LONG);
+            return true;
+        }
     }
-    if (l == 1) {
-        // 1 byte word: we index single ASCII letters and digits, but nothing else. We might want to
-        // turn this into a test for a single utf8 character instead ? TBD: the calling code should
-        // know the size in characters, so turning this to an Unicode test could be for free.
+    if (w.size() == 1) {
+        // 1 byte word: we index single ASCII letters and digits, but nothing else. Why?
         unsigned int c = ((unsigned int)w[0]) & 0xff;
         if (charclasses[c] != A_ULETTER && charclasses[c] != A_LLETTER && charclasses[c] != DIGIT &&
             (!(m_flags & TXTS_KEEPWILD) || charclasses[c] != WILD)  ) {
@@ -395,10 +457,10 @@ inline bool TextSplit::emitterm(bool isspan, const string& w, int pos, size_t bt
         }
     }
     // Detect doublons. Happens...?
-    if (pos != m_prevpos || l != m_prevlen) {
-        bool ret = takeword(w, pos, int(btstart), int(btend));
+    if (pos != m_prevpos || w.size() != m_prevlen) {
+        bool ret = takeword(w, pos, btstart, btend);
         m_prevpos = pos;
-        m_prevlen = int(w.length());
+        m_prevlen = w.size();
         return ret;
     }
     LOGDEB2("TextSplit::emitterm:dup: [" << w << "] pos " << pos << "\n");
@@ -510,16 +572,16 @@ inline bool TextSplit::doemit(bool spanerase, size_t _bp)
 
     if (m_wordLen) {
         // We have a current word. Remember it
-
-        if (int(m_words_in_span.size()) >= o_maxWordsInSpan) {
-            // Limit max span word count
-            spanerase = true;
-        }
-        
         if (!(o_noNumbers && m_inNumber)) {
             m_words_in_span.push_back({m_wordStart, m_wordStart + m_wordLen});
             m_wordpos++;
         }
+        if (int(m_words_in_span.size()) >= o_maxWordsInSpan) {
+            // Limit max span word count
+            spanerase = true;
+            discarded(m_span, m_spanpos, m_words_in_span[0].first, m_words_in_span.back().second,
+                      LONG_SPAN_TRUNCATED);
+        }
         m_wordLen = m_wordChars = 0;
     }
 
@@ -528,7 +590,7 @@ inline bool TextSplit::doemit(bool spanerase, size_t _bp)
         m_wordStart = int(m_span.length());
         return true;
     }
-
+    
     // Span is done (too long or span-terminating character). Produce terms and reset it.
     string initials;
     if (span_is_initials(initials)) {
@@ -608,16 +670,13 @@ bool TextSplit::text_to_words(const string &in)
     bool nlpending = false;
     bool softhyphenpending = false;
 
-    // Running count of non-alphanum chars. Reset when we see one;
-    int nonalnumcnt = 0;
-
     Utf8Iter it(in);
 #if defined(KATAKANA_AS_WORDS) || defined(HANGUL_AS_WORDS) || defined(CHINESE_AS_WORDS)
     int prev_csc = -1;
 #endif
+    int prevc = -1;
     for (; !it.eof() && !it.error(); it++) {
         unsigned int c = *it;
-        nonalnumcnt++;
 
         if (c == (unsigned int)-1) {
             LOGERR("Textsplit: error occurred while scanning UTF-8 string\n");
@@ -630,19 +689,25 @@ bool TextSplit::text_to_words(const string &in)
         // segment AND a specific processor (e.g. external Hangul word segmenter) has been
         // configured. In other words, CJK characters are processed by the generic ngram term
         // generator, except if a language-specific processor has been implemented and configured.
+        bool gospecial = true;
         if (UNICODE_IS_KATAKANA(c)) {
             csc = CSC_KATAKANA;
         } else if (UNICODE_IS_HANGUL(c)) {
             csc = CSC_HANGUL;
         } else if (UNICODE_IS_CHINESE(c)) {
             csc = CSC_CHINESE;
+        } else if (UNICODE_IS_OTHER_NGRAM(c)) {
+            csc = CSC_OTHERNGRAM;
+        } else if (UNICODE_IS_TESTNGRAM(c)) {
+            csc = CSC_TESTNGRAM;
         } else if (UNICODE_IS_CJK(c)) {
             csc = CSC_CJK;
         } else {
+            gospecial = false;
             csc = CSC_OTHER;
         }
-        if (o_processCJK && (csc == CSC_CJK || csc == CSC_HANGUL || csc == CSC_CHINESE)) {
-            // CJK character hit. Hangul and Chineseprocessing may be special.
+        if (o_processCJK && gospecial) {
+            // Using ngrams with possible exceptions for Hangul and Chinese if so configured.
 
             // Do like at EOF with the current non-cjk data.
             if (m_wordLen || m_span.length()) {
@@ -720,7 +785,6 @@ bool TextSplit::text_to_words(const string &in)
             continue;
 
         case DIGIT:
-            nonalnumcnt = 0;
             if (m_wordLen == 0)
                 m_inNumber = true;
             m_wordLen += it.appendchartostring(m_span);
@@ -728,7 +792,6 @@ bool TextSplit::text_to_words(const string &in)
             break;
 
         case SPACE:
-            nonalnumcnt = 0;
         SPACE:
             if (m_wordLen || m_span.length()) {
                 if (!doemit(true, it.getBpos()))
@@ -743,6 +806,16 @@ bool TextSplit::text_to_words(const string &in)
                 nlpending = false;
                 newline(m_wordpos);
             }
+
+            if (o_haveidxpunct && (o_idxpunct.find(c) != o_idxpunct.end())) {
+                std::string tmp;
+                auto l = it.appendchartostring(tmp);
+                // std::cerr << "IDXSPACE: [" << c << "] [" << tmp << "]\n";
+                takeword(tmp, m_wordpos, it.getBpos(), it.getBpos()+l);
+                if (m_spanpos == m_wordpos)
+                    m_spanpos++;
+                m_wordpos++;
+            }
             break;
 
         case WILD:
@@ -763,10 +836,9 @@ bool TextSplit::text_to_words(const string &in)
                     m_wordLen += it.appendchartostring(m_span);
                     STATS_INC_WORDCHARS;
                     break;
-                } 
-            } else if (m_inNumber) {
-                if ((m_span[m_span.length() - 1] == 'e' ||
-                     m_span[m_span.length() - 1] == 'E')) {
+                }
+            } else if (!o_numberDashSpans && m_inNumber) {
+                if ((m_span[m_span.length() - 1] == 'e' || m_span[m_span.length() - 1] == 'E')) {
                     if (isdigit(whatcc(it[it.getCpos()+1]), m_flags)) {
                         m_wordLen += it.appendchartostring(m_span);
                         STATS_INC_WORDCHARS;
@@ -776,8 +848,9 @@ bool TextSplit::text_to_words(const string &in)
             } else {
                 int nextc = it[it.getCpos()+1];
                 if (cc == '+') {
-                    if (nextc == '+' || nextc == -1 || isvisiblewhite(nextc)) {
-                        // someword++[+...] !
+                    if ((nextc == '+' && prevc != '+') ||
+                        (prevc == '+' && (nextc == -1 || isvisiblewhite(nextc)))) {
+                        // someword++ but not someword+ or someword+++
                         m_wordLen += it.appendchartostring(m_span);
                         STATS_INC_WORDCHARS;
                         break;
@@ -819,8 +892,13 @@ bool TextSplit::text_to_words(const string &in)
                 if (!isalphanum(nextwhat, m_flags)) {
                     // Do append the dot to the span: for abbrev recognition.
                     it.appendchartostring(m_span);
-                    // Set the iterator to point to the look-ahead char.
+                    // Set the iterator to point to the look-ahead char. Need to
+                    // test for nl/newpage here (there should be a better way?)
                     it++;
+                    if (*it == '\n')
+                        nlpending = true;
+                    else if (*it == '\f')
+                        pagepending = true;
                     goto SPACE;
                 }
 
@@ -857,8 +935,7 @@ bool TextSplit::text_to_words(const string &in)
         case '_': // If underscoreasletter is set, we'll never get this
         case '\'':
         {
-            // If in word, potential span: o'brien, jf@dockes.org,
-            // else just ignore
+            // If in word, potential span: o'brien, jf@dockes.org, else treat as SPACE
             int nextc = it[it.getCpos()+1];
             if (nextc == -1 || isvisiblewhite(nextc)) {
                 goto SPACE;
@@ -961,7 +1038,6 @@ bool TextSplit::text_to_words(const string &in)
 
         default:
         NORMALCHAR:
-            nonalnumcnt = 0;
             if (m_inNumber && c != 'e' && c != 'E') {
                 m_inNumber = false;
             }
@@ -970,6 +1046,7 @@ bool TextSplit::text_to_words(const string &in)
             break;
         }
         softhyphenpending = false;
+        prevc = cc;
     }
     if (m_wordLen || m_span.length()) {
         if (!doemit(true, it.getBpos()))
@@ -983,7 +1060,7 @@ class TextSplitCW : public TextSplit {
 public:
     int wcnt;
     TextSplitCW(int flags) : TextSplit(flags), wcnt(0) {}
-    bool takeword(const string &, int, int, int) {
+    bool takeword(const string &, size_t, size_t, size_t) override {
         wcnt++;
         return true;
     }
